#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sentence.h"
#include "ultrastarsong.h"

#define MAXCHARWORD 32

/***********************************************************************************//**
 * ctor
 **************************************************************************************/
cUltraStarSong::cUltraStarSong(  )
{
  m_lstSentences = NULL;
  Init(  );
}

/***********************************************************************************//**
 * dtor
 **************************************************************************************/
cUltraStarSong::~cUltraStarSong(  )
{
  Init( 0 );
}

/***********************************************************************************//**
 * reset member data, delete all sentence
 *
 * @param [in] realloc : 1: allocate sentence
 *                      0: don't allocate sentence
 **************************************************************************************/
void cUltraStarSong::Init( int realloc )
{
  m_nbMaxSentences = 0;
  m_nbSentences = 0;
  m_title[0] = 0;
  m_artiste[0] = 0;
  m_mp3FileName[0] = 0;
  m_BPM = 0;
  m_GAP = 0;
  m_relativeTime = 0;
  m_currentTime = 0;
  if( m_lstSentences != NULL )
  {
    delete[]m_lstSentences;
    m_lstSentences = NULL;
  }

  if( realloc == 1 )
  {
    m_lstSentences = new cSentence[1000];
    m_nbMaxSentences = 1000;
  }
}

/***********************************************************************************//**
 * read a ultraStar txt files
 *
 * @param [in] filename : file name to read
 **************************************************************************************/
void cUltraStarSong::ReadSong( const char *filename )
{
  FILE *file;
  char oneLine[256];
  char *ptrLine;
  Init(  );
  file = fopen( filename, "rt" );
  if( file == NULL )
  {
    fprintf( stderr, "Can't open file: '%s'", filename );
  }
  else
  {
    ptrLine = fgets( oneLine, 256, file );
    while( ptrLine != NULL )
    {
      if( oneLine[0] == '#' )   // header
      {
        ParseHeaderLine( oneLine );
      }
      else
      {
        ParseBodyLine( oneLine );
      }
      ptrLine = fgets( oneLine, 256, file );
    }
    fclose( file );

    /*
     * Add path to mp3 filename
     */
    MakeMp3FilePath( filename );
  }
}

/***********************************************************************************//**
 * generate the mp3 filename associte with the ultraStar txt file
 *
 * @param [in] filename : ultrastart file name
 **************************************************************************************/
void cUltraStarSong::MakeMp3FilePath( const char *filename )
{
  char *lastSlash;
  char tmp[256];
  strcpy( tmp, filename );
  lastSlash = strrchr( tmp, '/' );
  if( lastSlash != NULL )

  {
    lastSlash++;
    strcpy( lastSlash, m_mp3FileName );
    strcpy( m_mp3FileName, tmp );
  }
}

/***********************************************************************************//**
 * Trim Right
 *
 * @param [in/out] ptr : a string
 **************************************************************************************/
void cUltraStarSong::TrimR( char *ptr )
{
  char *stim;

  stim = strchr( ptr, 0x0a );
  if( stim != NULL )
    *stim = 0;
  stim = strchr( ptr, 0x0d );
  if( stim != NULL )
    *stim = 0;
}

/***********************************************************************************//**
 * parse a line of header
 *
 * @param [in] line : a string
 **************************************************************************************/
void cUltraStarSong::ParseHeaderLine( const char *line )
{
  if( strstr( line, "#TITLE:" ) != NULL )

  {
    strcpy( m_title, line + 7 );
    TrimR( m_title );
    printf( "[karaoke] Title:'%s'\n", m_title );
  }
  if( strstr( line, "#ARTIST:" ) != NULL )

  {
    strcpy( m_artiste, line + 8 );
    TrimR( m_artiste );
    printf( "[karaoke] Artist:'%s'\n", m_artiste );
  }
  if( strstr( line, "#MP3:" ) != NULL )

  {
    strcpy( m_mp3FileName, line + 5 );
    TrimR( m_mp3FileName );
    printf( "[karaoke] mp3 file:'%s'\n", m_mp3FileName );
  }
  if( strstr( line, "#BPM:" ) != NULL )

  {
    m_BPM = 4 * atoi( line + 5 );
    printf( "[karaoke] BPM:%d\n", m_BPM / 4 );
  }
  if( strstr( line, "#GAP:" ) != NULL )

  {
    m_GAP = atoi( line + 5 );
    printf( "[karaoke] GAP:%d\n", m_GAP );
  }
  if( strstr( line, "#RELATIVE:YES" ) != NULL )
  {
    m_relativeTime = 1;
    printf( "[karaoke] RELATIVE:yes\n" );
  }
}

/***********************************************************************************//**
 * parse a line of body
 *
 * @param [in] line : a string
 **************************************************************************************/
void cUltraStarSong::ParseBodyLine( char *line )
{
  char *pline = line;
  char strTmp[MAXCHARWORD];
  char strSyllable[MAXCHARSYLLABLE];
  int start, duration, note;

  if( line[0] == ':' || line[0] == '*' || line[0] == 'F' )
  {
    pline = GetLineParam( pline, strTmp, MAXCHARWORD );  // dummy
    pline = GetLineParam( pline, strTmp, MAXCHARWORD );  // Start
    start = atoi( strTmp );
    pline = GetLineParam( pline, strTmp, MAXCHARWORD );  // Duration
    duration = atoi( strTmp );
    pline = GetLineParam( pline, strTmp, MAXCHARWORD );  // Note
    note = atoi( strTmp );
    pline = GetLineParamTxt( pline, strSyllable, MAXCHARSYLLABLE );  // cSyllable

    if( m_relativeTime == 0 )
      m_currentTime = start;
    else
      m_currentTime += start;
    m_lstSentences[m_nbSentences].AddSyllable( m_currentTime, duration, note, strSyllable );
  }

  if( line[0] == '-' )
  {
    pline = GetLineParam( pline, strTmp, MAXCHARWORD );  // dummy
    pline = GetLineParam( pline, strTmp, MAXCHARWORD );  // Start
    start = atoi( strTmp );
    if( m_relativeTime == 0 )
      m_currentTime = start;
    else
      m_currentTime += start;
    if( m_nbSentences < m_nbMaxSentences )
    {
      m_nbSentences++;
      m_lstSentences[m_nbSentences].SetStart( start );
    }
  }
}

/***********************************************************************************//**
 * extract parameters
 *
 * @param [in] line :
 * @param [out] out :
 * @param [in] lengthMax :
 **************************************************************************************/
char *cUltraStarSong::GetLineParam( char *line, char *out, int lengthMax )
{
  while( ( *line ) != 0 && ( *line ) != ' ' && ( *line ) != 0xd && ( *line ) != 0xa )
  {
    *out = *line;
    line++;
    out++;
  }
  if( ( *line ) == ' ' )
    line++;
  *out = 0;
  return ( line );
}

/***********************************************************************************//**
 * extract parameters
 *
 * @param [in] line :
 * @param [out] out :
 * @param [in] lengthMax :
 **************************************************************************************/
char *cUltraStarSong::GetLineParamTxt( char *line, char *out, int lengthMax )
{
  while( ( *line ) != 0 && ( *line ) != 0xd && ( *line ) != 0xa )
  {
    *out = *line;
    line++;
    out++;
  }
  *out = 0;
  return ( line );
}

/***********************************************************************************//**
 * define the current time position in the song
 *
 * @param [in] float :
 **************************************************************************************/
int cUltraStarSong::SetCurrentTime( float time )
{
  int fin = 0;
  time -= m_GAP / 1000;
  time = ( time * m_BPM ) / 60;
  m_currentTime = ( int ) time;
  m_currentSentence = 0;
  while( fin == 0 )

  {
    if( m_currentTime <= m_lstSentences[m_currentSentence + 1].m_start )

    {
      m_lstSentences[m_currentSentence].SetAtBegin(  );
      fin = 1;
    }

    else
      m_currentSentence++;
    if( m_currentSentence >= m_nbSentences )
      fin = 1;
  }
  return ( 0 );
}

/***********************************************************************************//**
 * get next syllable in a sentence
 *
 * @param [out] syll :
 * @param [in] status :
 *
 * @return :
 **************************************************************************************/
int cUltraStarSong::GetNextSyllable( cSyllable & syll, int *status )
{
  syll.Erase();
  if( m_lstSentences[m_currentSentence].GetNextSyllable( syll, status, m_currentTime ) == 1 )
  {
    m_currentSentence++;
    m_lstSentences[m_currentSentence].SetAtBegin(  );
  }
  return ( m_currentSentence );
}
